from typing import Dict
import jax
import jax.numpy as jnp

import navix as nx
from navix.components import EMPTY_POCKET_ID, DISCARD_PILE_COORDS
from navix.entities import Entities, Entity, Directions
from navix.states import State
from navix.rendering.registry import PALETTE


def test_rotation():
    direction = jnp.asarray(0)

    key = jax.random.PRNGKey(0)
    grid = jnp.zeros((3, 3), dtype=jnp.int32)
    player = nx.entities.Player(
        position=jnp.asarray((1, 1)), direction=direction, pocket=EMPTY_POCKET_ID
    )[None]
    cache = nx.rendering.cache.RenderingCache.init(grid)

    entities: Dict[str, Entity] = {
        Entities.PLAYER: player,
    }

    player.check_ndim(batched=True)

    state = State(
        grid=grid,
        entities=entities,
        cache=cache,
        key=key,
    )

    msg = "Expected direction to be {}, got {}"
    state = nx.actions._rotate(state, -1)
    player = state.get_player()
    player.check_ndim(batched=False)
    assert player.direction == jnp.asarray(3), msg.format(3, player.direction)

    state = nx.actions._rotate(state, 1)
    player = state.get_player()
    player.check_ndim(batched=False)
    assert player.direction == jnp.asarray(0), msg.format(0, player.direction)

    state = nx.actions._rotate(state, 2)
    player = state.get_player()
    player.check_ndim(batched=False)
    assert player.direction == jnp.asarray(2), msg.format(2, player.direction)

    state = nx.actions._rotate(state, 3)
    player = state.get_player()
    player.check_ndim(batched=False)
    assert player.direction == jnp.asarray(1), msg.format(1, player.direction)
    return


def test_move():
    heigh, width = 5, 5
    grid = jnp.zeros((heigh - 2, width - 2), dtype=jnp.int32)
    grid = jnp.pad(grid, pad_width=1, mode="constant", constant_values=1)
    key = jax.random.PRNGKey(0)
    player = nx.entities.Player(
        position=jnp.asarray((1, 1)), direction=jnp.asarray(0), pocket=EMPTY_POCKET_ID
    )
    goals = nx.entities.Goal.create(
        position=jnp.asarray((3, 3)), probability=jnp.asarray(1.0)
    )
    keys = nx.entities.Key(
        position=jnp.asarray((3, 1)), id=jnp.asarray(-1), colour=PALETTE.YELLOW
    )
    doors = nx.entities.Door(
        position=jnp.asarray((2, 2)),
        requires=jnp.asarray(-1),
        open=jnp.asarray(False),
        colour=PALETTE.YELLOW,
    )
    cache = nx.rendering.cache.RenderingCache.init(grid)

    player.check_ndim(batched=False)
    goals.check_ndim(batched=False)
    keys.check_ndim(batched=False)
    doors.check_ndim(batched=False)

    """
    #  #  #  #  #
    #  P  .  . #
    #  .  D  . #
    #  K  .  G #
    #  #  #  #  #
    """
    entities = {
        Entities.PLAYER: player[None],
        Entities.GOAL: goals[None],
        Entities.KEY: keys[None],
        Entities.DOOR: doors[None],
    }
    state = State(
        key=key,
        grid=grid,
        entities=entities,
        cache=cache,
    )

    # check forward
    state = nx.actions.forward(state)
    expected_position = jnp.asarray((1, 2))
    player = state.get_player()
    player.check_ndim(batched=False)
    assert jnp.array_equal(player.position, expected_position)

    # check backward
    state = nx.actions.backward(state)
    expected_position = jnp.asarray((1, 1))
    player = state.get_player()
    player.check_ndim(batched=False)
    assert jnp.array_equal(player.position, expected_position)

    # check right
    state = nx.actions.right(state)
    expected_position = jnp.asarray((2, 1))
    player = state.get_player()
    player.check_ndim(batched=False)
    assert jnp.array_equal(player.position, expected_position)

    # check left
    state = nx.actions.left(state)
    expected_position = jnp.asarray((1, 1))
    player = state.get_player()
    player.check_ndim(batched=False)
    assert jnp.array_equal(player.position, expected_position)

    # check that we can't walk through a closed door
    state = nx.actions.forward(state)
    state = nx.actions.rotate_cw(state)
    state = nx.actions.forward(state)
    expected_position = jnp.asarray((1, 2))
    player = state.get_player()
    player.check_ndim(batched=False)
    assert jnp.array_equal(player.position, expected_position)

    # check that we can walk through an open door
    state.entities["door"] = entities["door"].replace(open=jnp.asarray(True)[None])
    state = nx.actions.forward(state)
    state = nx.actions.forward(state)
    expected_position = jnp.asarray((3, 2))
    player = state.get_player()
    player.check_ndim(batched=False)
    assert jnp.array_equal(player.position, expected_position)

    # check that we can walk through a door backwards
    state = nx.actions.backward(state)
    state = nx.actions.backward(state)
    expected_position = jnp.asarray((1, 2))
    player = state.get_player()
    player.check_ndim(batched=False)
    assert jnp.array_equal(player.position, expected_position)


def test_walkable():
    height, width = 5, 5
    grid = nx.grid.room(height, width)
    key = jax.random.PRNGKey(0)
    player = nx.entities.Player(
        position=jnp.asarray((1, 1)), direction=Directions.EAST, pocket=EMPTY_POCKET_ID
    )
    goals = nx.entities.Goal.create(
        position=jnp.asarray((3, 3)), probability=jnp.asarray(1.0)
    )
    keys = nx.entities.Key(
        position=jnp.asarray((3, 1)), id=jnp.asarray(1), colour=PALETTE.YELLOW
    )
    doors = nx.entities.Door(
        position=jnp.asarray((1, 3)),
        requires=jnp.asarray(1),
        open=jnp.asarray(False),
        colour=PALETTE.YELLOW,
    )
    cache = nx.rendering.cache.RenderingCache.init(grid)

    player.check_ndim(batched=False)
    goals.check_ndim(batched=False)
    keys.check_ndim(batched=False)
    doors.check_ndim(batched=False)

    # Looks like this
    """
    #  #  #  #  #
    #  P  .  D #
    #  .  .  . #
    #  K  .  G #
    #  #  #  #  #
    """
    state = State(
        key=key,
        grid=grid,
        cache=cache,
        entities={
            Entities.PLAYER: player[None],
            Entities.GOAL: goals[None],
            Entities.KEY: keys[None],
            Entities.DOOR: doors[None],
        },
    )

    def check_forward_position(state: State):
        player = state.get_player()
        player.check_ndim(batched=False)
        prev_pos = player.position
        pos = nx.grid.translate_forward(prev_pos, player.direction, jnp.asarray(1))
        walkable, _ = nx.actions._can_walk_there(state, pos)
        assert (
            not walkable
        ), "Expected position {} to be not walkable, since it is a {}".format(
            pos, state.grid[tuple(pos)]
        )

    # should not be able to walk on/through a closed door
    state = nx.actions.forward(state)
    check_forward_position(state)

    # should be able to walk through an open door
    doors = state.get_doors()
    state.entities[Entities.DOOR] = doors.replace(open=jnp.asarray(True)[None])
    state = nx.actions.forward(state)
    state = nx.actions.backward(state)

    # should not be able to walk on/through a key
    state = nx.actions.backward(state)
    state = nx.actions.rotate_cw(state)
    state = nx.actions.forward(state)
    check_forward_position(state)

    # should not be able to walk on/through a wall
    state = nx.actions.rotate_ccw(state)
    state = nx.actions.forward(state)
    state = nx.actions.rotate_cw(state)
    state = nx.actions.forward(state)
    check_forward_position(state)

    # should be able to walk on/through a goal
    state = nx.actions.rotate_ccw(state)

    # but not on/through a wall
    state = nx.actions.forward(state)
    check_forward_position(state)


def test_pickup():
    heigh, width = 5, 5
    grid = jnp.zeros((heigh - 2, width - 2), dtype=jnp.int32)
    grid = jnp.pad(grid, pad_width=1, mode="constant", constant_values=1)
    key = jax.random.PRNGKey(0)
    player = nx.entities.Player(
        position=jnp.asarray((1, 1)), direction=jnp.asarray(1), pocket=EMPTY_POCKET_ID
    )
    goals = nx.entities.Goal.create(
        position=jnp.asarray((3, 3)), probability=jnp.asarray(1.0)
    )
    keys = nx.entities.Key(
        position=jnp.asarray((2, 1)), id=jnp.asarray(1), colour=PALETTE.YELLOW
    )
    doors = nx.entities.Door(
        position=jnp.asarray((1, 3)),
        requires=jnp.asarray(1),
        open=jnp.asarray(False),
        colour=PALETTE.YELLOW,
    )
    cache = nx.rendering.cache.RenderingCache.init(grid)

    # Looks like this
    """
    #  #  #  #  #
    #  P  .  D  #
    #  K  .  .  #
    #  .  .  G  #
    #  #  #  #  #
    """
    entities = {
        Entities.PLAYER: player[None],
        Entities.GOAL: goals[None],
        Entities.KEY: keys[None],
        Entities.DOOR: doors[None],
    }
    state = State(
        key=key,
        grid=grid,
        cache=cache,
        entities=entities,
    )

    # check that the player has no keys
    player = state.get_player()
    player.check_ndim(batched=False)
    assert jnp.array_equal(
        player.pocket, EMPTY_POCKET_ID
    ), "Expected player to have pocket {}, got {}".format(
        EMPTY_POCKET_ID, player.pocket
    )

    # pick up the key
    state = nx.actions.pickup(state)

    # check that the player has the key
    player = state.get_player()
    player.check_ndim(batched=False)
    expected_pocket = jnp.asarray(1)
    assert jnp.array_equal(
        player.pocket, expected_pocket
    ), "Expected player to have key {}, got {}".format(expected_pocket, player.pocket)

    # check that the key is no longer on the grid
    keys = state.get_keys()
    assert jnp.array_equal(
        keys.position[0], DISCARD_PILE_COORDS
    ), "Expected key to be at {}, got {}".format(DISCARD_PILE_COORDS, keys.position)


def test_open():
    heigh, width = 5, 5
    grid = jnp.zeros((heigh - 2, width - 2), dtype=jnp.int32)
    grid = jnp.pad(grid, pad_width=1, mode="constant", constant_values=1)
    key = jax.random.PRNGKey(0)
    player = nx.entities.Player(
        position=jnp.asarray((1, 1)), direction=jnp.asarray(0), pocket=EMPTY_POCKET_ID
    )
    goals = nx.entities.Goal.create(
        position=jnp.asarray((3, 3)), probability=jnp.asarray(1.0)
    )
    keys = nx.entities.Key(
        position=jnp.asarray((3, 1)), id=jnp.asarray(1), colour=PALETTE.YELLOW
    )
    doors = nx.entities.Door(
        position=jnp.asarray((1, 3)),
        requires=jnp.asarray(1),
        open=jnp.asarray(False),
        colour=PALETTE.YELLOW,
    )
    cache = nx.rendering.cache.RenderingCache.init(grid)

    player.check_ndim(batched=False)
    goals.check_ndim(batched=False)
    keys.check_ndim(batched=False)
    doors.check_ndim(batched=False)

    # Looks like this
    # W  W  W  W  W
    # W  P  0  D  W
    # W  0  0  0  W
    # W  K  0  G  W
    # W  W  W  W  W
    entities = {
        Entities.PLAYER: player[None],
        Entities.GOAL: goals[None],
        Entities.KEY: keys[None],
        Entities.DOOR: doors[None],
    }

    state = State(
        key=key,
        grid=grid,
        cache=cache,
        entities=entities,
    )

    # check that the player has no keys
    player = state.get_player()
    player.check_ndim(batched=False)
    expected_pocket = EMPTY_POCKET_ID
    assert jnp.array_equal(
        player.pocket, expected_pocket
    ), "Expected player to have {}, got {}".format(expected_pocket, player.pocket)

    state = nx.actions.forward(state)

    # check that pocket is empty
    player = state.get_player()
    player.check_ndim(batched=False)
    expected_pocket = EMPTY_POCKET_ID
    assert jnp.array_equal(
        player.pocket, expected_pocket
    ), "Expected player to have pocket {}, got {}".format(
        expected_pocket, player.pocket
    )

    # check that we cannot open a door without the required key
    state = nx.actions.open(state)
    # and that the door is still closed
    doors = state.get_doors()
    doors.check_ndim(batched=True)
    expected_open = jnp.asarray(False)[None]
    assert jnp.array_equal(
        doors.open, expected_open
    ), "Expected door open status {}, got {}".format(expected_open, doors.open)

    # artificially put the right key in the player's pocket
    player = state.get_player()
    player.check_ndim(batched=False)
    player = player.replace(pocket=jnp.asarray(1))
    player.check_ndim(batched=False)
    state = state.set_player(player)
    expected_pocket = jnp.asarray(1)
    assert jnp.array_equal(
        player.pocket, expected_pocket
    ), "Expected player to have pocket {}, got {}".format(
        expected_pocket, player.pocket
    )

    # check that we can open the door with the right key
    state = nx.actions.open(state)
    doors = state.get_doors()
    doors.check_ndim(batched=True)
    expected_open = jnp.asarray(True)[None]
    assert jnp.array_equal(
        doors.open, expected_open
    ), "Expected door open status {}, got {}".format(expected_open, doors.open)

    # check that opening an open door keeps it open
    state = nx.actions.open(state)
    doors = state.get_doors()
    doors.check_ndim(batched=True)
    expected_open = jnp.asarray(True)[None]
    assert jnp.array_equal(
        doors.open, expected_open
    ), "Expected door open status {}, got {}".format(expected_open, doors.open)

    # check that we can walk through an open door
    state = nx.actions.forward(state)
    player = state.get_player()
    player.check_ndim(batched=False)
    expected_position = jnp.asarray((1, 3))
    assert jnp.array_equal(
        player.position, expected_position
    ), "Expected player position to be {}, got {}".format(
        expected_position, player.position
    )


if __name__ == "__main__":
    # test_rotation()
    # test_move()
    test_walkable()
    test_pickup()
    # test_open()



from dataclasses import dataclass, field
from typing import Dict

import distrax
import tyro
import numpy as np
import jax.numpy as jnp
import navix as nx
from navix import observations
from navix.agents import PPO, PPOHparams, ActorCritic
from navix.environments.environment import Environment


@dataclass
class Args:
    project_name = "navix-debug"
    population_size: int = 10
    seed: int = 0
    # env
    env_id: str = "Navix-DoorKey-Random-6x6-v0"
    discount: float = 0.99
    # ppo
    ppo_config: PPOHparams = field(default_factory=PPOHparams)


class CategoricalUniform(distrax.Categorical):
    def __init__(self, domain: tuple, dtype=jnp.int32):
        self.domain = jnp.asarray(domain)
        super().__init__(logits=jnp.zeros(len(domain)), dtype=dtype)

    def sample(self, *, seed, sample_shape=()):
        samples = super().sample(seed=seed, sample_shape=sample_shape)
        return self.domain[samples]

    def sample_n(self, rng, n):
        samples = super().sample(seed=rng, sample_shape=(n,))
        return self.domain[samples]


if __name__ == "__main__":
    args = tyro.cli(Args)

    def FlattenObsWrapper(env: Environment):
        flatten_obs_fn = lambda x: jnp.ravel(env.observation_fn(x))
        flatten_obs_shape = (int(np.prod(env.observation_space.shape)),)
        return env.replace(
            observation_fn=flatten_obs_fn,
            observation_space=env.observation_space.replace(shape=flatten_obs_shape),
        )

    env = nx.make(
        args.env_id,
        observation_fn=observations.symbolic,
        gamma=args.discount,
    )
    env = FlattenObsWrapper(env)

    # static hparams
    ppo_config = args.ppo_config.replace(anneal_lr=False)

    hparams_distr: Dict[str, distrax.Distribution] = {
        "gae_lambda": CategoricalUniform((0.7, 0.95, 0.99)),
        "clip_eps": CategoricalUniform((0.1, 0.2)),
        "ent_coef": CategoricalUniform((0.001, 0.01, 0.1)),
        "vf_coef": CategoricalUniform((0.1, 0.5, 0.9)),
        "lr": CategoricalUniform((1e-3, 2.5e-4, 1e-4, 1e-5)),
    }

    base_hparams = args.ppo_config
    experiment = nx.Experiment(
        name=args.project_name,
        agent=PPO(base_hparams, ActorCritic(len(env.action_set)), env),
        env=env,
        env_id=args.env_id,
        seeds=(args.seed,),
    )

    experiment.run_hparam_search(hparams_distr, args.population_size)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from jax import Array
import jax.numpy as jnp

from .states import State
from .grid import positions_equal, translate
from .entities import Entities, Player


def on_goal_reached(state: State) -> Array:
    """Checks whether the goal has been reached using the `goal_reached` event.

    Args:
        state (State): The current state of the game.

    Returns:
        Array: A boolean array indicating whether the goal has been reached."""
    return state.events.goal_reached.happened


def on_lava_fall(state: State) -> Array:
    """Checks whether the lava has fallen using the `lava_fall` event.

    Args:
        state (State): The current state of the game.

    Returns:
        Array: A boolean array indicating whether the lava has fallen."""
    return state.events.lava_fall.happened


def on_ball_hit(state: State) -> Array:
    """Checks whether the ball has hit something using the `ball_hit` event.

    Args:
        state (State): The current state of the game.

    Returns:
        Array: A boolean array indicating whether the ball has hit something."""
    return state.events.ball_hit.happened


def on_door_done(state: State) -> Array:
    """Checks whether the action `done` has been called in front of a `Door` object with the correct colour.

    Args:
        state (State): The current state of the game.

    Returns:
        Array: A boolean array indicating whether the action `done` has been called in front of a `Door` object with the correct colour.
    """
    assert (
        state.mission is not None
    ), "Termination on door done requires the state to specify a mission."
    player = state.entities[Entities.PLAYER][0]
    assert isinstance(player, Player)

    fwd_pos = translate(player.position, player.direction)
    if Entities.DOOR not in state.entities:
        return jnp.asarray(False)
    doors = state.get_doors()
    idx = jnp.where(positions_equal(doors.position, fwd_pos), size=1)[0][0]
    doors = doors[idx]
    pos_match = jnp.array_equal(fwd_pos, state.mission.position)
    colour_match = jnp.array_equal(doors.colour, state.mission.colour)
    return jnp.logical_and(pos_match, colour_match)


def on_wall_hit(state: State) -> Array:
    """Checks whether the wall has been hit using the `wall_hit` event.

    Args:
        state (State): The current state of the game.

    Returns:
        Array: A boolean array indicating whether the wall has been hit."""
    return state.events.wall_hit.happened



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations
from functools import partial


from typing import Callable, Dict, List, Tuple
import jax
import jax.numpy as jnp
from jax import Array
import jax.tree_util as jtu
from flax import struct


Coordinates = Tuple[Array, Array]


def coordinates(grid: Array) -> Coordinates:
    """Returns a tuple of 2D coordinates [(col, row), ...] for each cell in the grid.
    A grid array of shape `i32[height, width]` will return a tuple of length (height * width),
    containing two arrays, each of shape `i32[2]`.

    Args:
        grid (Array): A 2D grid of shape (height, width).
    
    Returns:
        Tuple[Array, Array]: A tuple of two arrays containing the 2D coordinates of \
        each cell in the grid.
    """
    return tuple(jnp.mgrid[0 : grid.shape[0], 0 : grid.shape[1]])  # type: ignore


def idx_from_coordinates(grid: Array, coordinates: Array) -> Array:
    """Converts a batch of 2D coordinates [(col, row), ...] into a flat index

    Args:
        grid (Array): A 2D grid of shape (height, width).
        coordinates (Array): A batch of 2D coordinates of shape (batch_size, 2).

    Returns:
        Array: A flat index of shape `i32[batch_size]` for each coordinate in the batch.
    """
    coordinates = coordinates.T
    assert coordinates.shape[0] == 2, coordinates.shape

    idx = coordinates[0] * grid.shape[1] + coordinates[1]
    return jnp.asarray(idx, dtype=jnp.int32)


def coordinates_from_idx(grid: Array, idx: Array) -> Array:
    """Converts a flat index of shape `i32[]` into a 2D coordinate `i32[2]` containing \
    (col, row) data. The index is calculated as `idx = row * width + col`.
    
    Args:
        grid (Array): A 2D grid of shape (height, width).
        idx (Array): A flat index of shape `i32[]`.
        
    Returns:
        Array: A 2D coordinate of shape `i32[2]` containing the (col, row) data."""
    coords = jnp.divmod(idx, grid.shape[1])
    return jnp.asarray(coords, dtype=jnp.int32).T


def mask_by_coordinates(
    grid: Array,
    address: Coordinates,
    comparison_fn: Callable[[Array, Array], Array] = jnp.greater_equal,
) -> Array:
    """This is a workaround to compute dynamicly-sized masks in XLA,
    which would not be possible otherwise.
    Returns a mask of the same shape as `grid` where the value is 1 if the
    corresponding element in `grid` satisfies the `comparison_fn` with the
    corresponding element in `address` (col, row) and 0 otherwise.

    Args:
        grid (Array): A 2D grid of shape (height, width).
        address (Coordinates): A tuple of 2D coordinates (col, row).
        comparison_fn (Callable[[Array, Array], Array], optional): A comparison function. \
        Defaults to `jnp.greater_equal`.

    Returns:
        Array: A boolean mask of the same shape as `grid`.

    """
    mesh = jnp.mgrid[0 : grid.shape[0], 0 : grid.shape[1]]
    cond_1 = comparison_fn(mesh[0], address[0])
    cond_2 = comparison_fn(mesh[1], address[1])
    mask = jnp.asarray(jnp.logical_and(cond_1, cond_2), dtype=jnp.int32)
    return mask


def translate(
    position: Array, direction: Array, modulus: Array = jnp.asarray(1)
) -> Array:
    """Translates a point in a grid by a given direction and modulus.
    
    Args:
        position (Array): A 2D coordinate of shape `i32[2]` containing the (col, row) data.
        direction (Array): A direction in the range [0, 1, 2, 3] representing the \
        cardinal directions [east, south, west, north].
        modulus (Array, optional): The modulus of the translation. Defaults to jnp.asarray(1).

    Returns:
        Array: A 2D coordinate of shape `i32[2]` containing the (col, row) data.
        """
    moves = (
        lambda position: position + jnp.asarray((0, modulus)),  # east
        lambda position: position + jnp.asarray((modulus, 0)),  # south
        lambda position: position + jnp.asarray((0, -modulus)),  # west
        lambda position: position + jnp.asarray((-modulus, 0)),  # north
    )
    return jax.lax.switch(direction, moves, position)


def translate_forward(position: Array, forward_direction: Array, modulus: Array):
    """Translates a point in a grid by a given forward direction and modulus.
    
    Args:
        position (Array): A 2D coordinate of shape `i32[2]` containing the (col, row) data.
        forward_direction (Array): A direction in the range [0, 1, 2, 3] representing the \
        cardinal directions [east, south, west, north].
        modulus (Array): The modulus of the translation.
    
    Returns:
        Array: A 2D coordinate of shape `i32[2]` containing the (col, row) data."""
    return translate(position, forward_direction, modulus)


def translate_left(position: Array, forward_direction: Array, modulus: Array):
    """Translates a point in a grid by a given left direction and modulus.
    
    Args:
        position (Array): A 2D coordinate of shape `i32[2]` containing the (col, row) data.
        forward_direction (Array): A direction in the range [0, 1, 2, 3] representing the \
        cardinal directions [east, south, west, north].
        modulus (Array): The modulus of the translation.
    
    Returns:
        Array: A 2D coordinate of shape `i32[2]` containing the (col, row) data."""
    return translate(position, (forward_direction + 3) % 4, modulus)


def translate_right(position: Array, forward_direction: Array, modulus: Array):
    """Translates a point in a grid by a given right direction and modulus.
    
    Args:
        position (Array): A 2D coordinate of shape `i32[2]` containing the (col, row) data.
        forward_direction (Array): A direction in the range [0, 1, 2, 3] representing the \
        cardinal directions [east, south, west, north].
        modulus (Array): The modulus of the translation.
        
    Returns:
        Array: A 2D coordinate of shape `i32[2]` containing the (col, row) data."""
    return translate(position, (forward_direction + 1) % 4, modulus)


def rotate(direction: Array, spin: int) -> Array:
    """Changes a direction vectory by a given number of spins.
    
    Args:
        direction (Array): A direction vector of shape `i32[]` in the range [0, 3] \
        representing the cardinal directions [east, south, west, north].
        spin (int): The number of spins to apply.
        
    Returns:
        Array: A direction vector of shape `i32[]` in the range [0, 3] representing \
        the cardinal directions [east, south, west, north]."""
    return (direction + spin) % 4


def align(patch: Array, current_direction: Array, desired_direction: Array) -> Array:
    """Aligns a patch of the grid from the current direction to the desired direction.
    
    Args:
        patch (Array): A patch of the grid.
        current_direction (Array): The current direction in the range [0, 1, 2, 3] \
        representing the cardinal directions [east, south, west, north].
        desired_direction (Array): The desired direction in the range [0, 1, 2, 3] \
        representing the cardinal directions [east, south, west, north].
        
    Returns:
        Array: A patch of the grid aligned to the desired direction."""
    return jax.lax.switch(
        desired_direction - current_direction,
        (
            lambda x: jnp.rot90(x, 1),  # 0 = transpose, 1 = flip
            lambda x: jnp.rot90(x, 2),  # 0 = flip, 1 = flip
            lambda x: jnp.rot90(x, 3),  # 0 = flip, 1 = transpose
            lambda x: x,
        ),
        patch,
    )


def random_positions(
    key: Array, grid: Array, n: int = 1, exclude: Array = jnp.asarray((-1, -1))
) -> Array:
    """Generates `n` random positions in the grid, excluding the `exclude` position.

    Args:
        key (Array): A random key.
        grid (Array): A 2D grid of shape (height, width).
        n (int, optional): The number of random positions to generate. Defaults to 1.
        exclude (Array, optional): The position to exclude. Defaults to jnp.asarray((-1, -1)).

    Returns:
        Array: A batch of random positions of shape `i32[n, 2]`."""
    probs = grid.reshape(-1)
    indices = idx_from_coordinates(grid, exclude)
    probs = probs.at[indices].set(-1) + 1.0
    idx = jax.random.categorical(key, jnp.log(probs), shape=(n,))
    position = coordinates_from_idx(grid, idx)
    return position.squeeze()


def random_directions(key: Array, n=1) -> Array:
    """Generates `n` random directions in the range [0, 1, 2, 3] representing the \
        cardinal directions [east, south, west, north].
        
    Args:
        key (Array): A random key.
        n (int, optional): The number of random directions to generate. Defaults to 1.
        
    Returns:
        Array: A batch of random directions of shape `i32[n]`."""
    return jax.random.randint(key, (n,), 0, 4).squeeze()


def random_colour(key: Array, n=1) -> Array:
    """Generates `n` random colours in the range [0, 1, 2, 3, 4, 5].

    Args:
        key (Array): A random key.
        n (int, optional): The number of random colours to generate. Defaults to 1.

    Returns:
        Array: A batch of random colours of shape `u8[n]`."""
    return jax.random.randint(key, (n,), 0, 6).squeeze()


def positions_equal(a: Array, b: Array) -> Array:
    """Checks if two points are equal.

    Args:
        a (Array): A 2D coordinate of shape `i32[2]` containing the (col, row) data.
        b (Array): A 2D coordinate of shape `i32[2]` containing the (col, row) data.

    Returns:

    """
    if b.ndim == 1:
        b = b[None]
    if a.ndim == 1:
        a = a[None]
    assert a.ndim == b.ndim == 2, (a.shape, b.shape)
    is_equal = jnp.all(jnp.equal(a, b), axis=-1)
    assert is_equal.shape == (max(b.shape[0], a.shape[0]),)
    return is_equal


def room(height: int, width: int) -> Array:
    """Creates an array representing a room of size `height` x `width`, including
    a set of walls around the room. The room is represented as a 2D grid of shape
    `(height, width)`, including walls, with walls set to -1 and empty tiles set to 0.

    Args:
        height (int): The height of the room.
        width (int): The width of the room.

    Returns:
        Array: A 2D grid of shape `(height, width)` representing a room."""
    grid = jnp.zeros((height - 2, width - 2), dtype=jnp.int32)
    return jnp.pad(grid, 1, mode="constant", constant_values=-1)


def two_rooms(height: int, width: int, key: Array) -> Tuple[Array, Array]:
    """Creates a 2D grid representing two rooms of size `height` x `width`, separated
    by a wall. The rooms are represented as a 2D grid of shape `(height, width)`, \
    including walls, with walls set to -1 and empty tiles set to 0.

    Args:
        height (int): The height of the rooms.
        width (int): The width of the rooms.
        key (Array): A random key, determining the position of the wall separating the rooms.
    
    Returns:
        Tuple[Array, Array]: A tuple containing the 2D grid representing the rooms \
        and the column index of the wall separating the rooms."""
    # create room
    grid = jnp.zeros((height - 2, width - 2), dtype=jnp.int32)
    grid = jnp.pad(grid, 1, mode="constant", constant_values=-1)

    # add separation wall
    wall_at = jax.random.randint(key, (), 2, width - 2)
    grid = grid.at[1:-1, wall_at].set(-1)
    return grid, wall_at


def vertical_wall(
    grid: Array, row_idx: int, opening_col_idx: Array | None = None
) -> Array:
    """Creates a vertical wall in the grid at the given row index, with an opening at the \
        given column index.
    
    Args:
        grid (Array): A 2D grid of shape `(height, width)`.
        row_idx (int): The row index where the wall is placed.
        opening_col_idx (Array, optional): The column index where the opening is placed. \
        Defaults to None.
    
    Returns:
        Array: A 2D grid of shape `(height, width)` with a vertical wall."""
    rows = jnp.arange(1, grid.shape[0] - 1)
    cols = jnp.asarray([row_idx] * (grid.shape[0] - 2))
    positions = jnp.stack((rows, cols), axis=1)
    if opening_col_idx is not None:
        positions = jnp.delete(
            positions, opening_col_idx - 1, axis=0, assume_unique_indices=True
        )
    return positions


def horizontal_wall(
    grid: Array, col_idx: int, opening_row_idx: Array | None = None
) -> Array:
    """Creates a horizontal wall in the grid at the given column index, with an opening at the \
        given row index.
        
    Args:
        grid (Array): A 2D grid of shape `(height, width)`.
        col_idx (int): The column index where the wall is placed.
        opening_row_idx (Array, optional): The row index where the opening is placed. \
        Defaults to None.
    
    Returns:
        Array: A 2D grid of shape `(height, width)` with a horizontal wall."""
    rows = jnp.asarray([col_idx] * (grid.shape[1] - 2))
    cols = jnp.arange(1, grid.shape[1] - 1)
    positions = jnp.stack((rows, cols), axis=1)
    if opening_row_idx is not None:
        positions = jnp.delete(
            positions, opening_row_idx - 1, axis=0, assume_unique_indices=True
        )
    return positions


def crop(
    grid: Array, origin: Array, direction: Array, radius: int, padding_value: int = 0
) -> Array:
    """Crops a grid around a given origin, facing a given direction, with a given radius.

    Args:
        grid (Array): A 2D grid of shape `(height, width)`.
        origin (Array): The origin of the crop.
        direction (Array): The direction the crop is facing.
        radius (int): The radius of the crop.
        padding_value (int, optional): The padding value. Defaults to 0.

    Returns:
        Array: A cropped grid."""
    input_shape = grid.shape
    # assert radius % 2, "Radius must be an odd number"
    # mid = jnp.asarray([g // 2 for g in grid.shape[:2]])
    # translated = jnp.roll(grid, mid - origin, axis=(0, 1))

    # # crop such that the agent is in the centre of the grid
    # cropped = translated.at[: 2 * radius + 1, : 2 * radius + 1].get(
    #     fill_value=padding_value
    # )

    # pad with radius
    padding = [(radius, radius), (radius, radius)]
    for _ in range(len(input_shape) - 2):
        padding.append((0, 0))

    padded = jnp.pad(grid, padding, constant_values=padding_value)

    # translate the grid such that the agent is `radius` away from the top and left edges
    translated = jnp.roll(padded, -jnp.asarray(origin), axis=(0, 1))

    # crop such that the agent is in the centre of the grid
    cropped = translated[: 2 * radius + 1, : 2 * radius + 1]

    # rotate such that the agent is facing north
    rotated = jax.lax.switch(
        direction,
        (
            lambda x: jnp.rot90(x, 1),  # 0 = transpose, 1 = flip
            lambda x: jnp.rot90(x, 2),  # 0 = flip, 1 = flip
            lambda x: jnp.rot90(x, 3),  # 0 = flip, 1 = transpose
            lambda x: x,
        ),
        cropped,
    )

    cropped = rotated.at[: radius + 1].get(fill_value=padding_value)
    return jnp.asarray(cropped, dtype=grid.dtype)


def view_cone(transparency_map: Array, origin: Array, radius: int) -> Array:
    """Computes the view cone of a given origin in a grid with a given radius.
    The view cone is a boolean map of transparent (1) and opaque (0) tiles, indicating
    whether a tile is visible from the origin or not.

    Args:
        transparency_map (Array): A boolean map of transparent (1) and opaque (0) tiles.
        origin (Array): The origin of the view cone.
        radius (int): The radius of the view cone.

    Returns:
        Array: The view cone of the given origin in the grid with the given radius."""
    # transparency_map is a boolean map of transparent (1) and opaque (0) tiles

    def fin_diff(array, _):
        array = jnp.roll(array, -1, axis=0) + array + jnp.roll(array, +1, axis=0)
        array = jnp.roll(array, -1, axis=1) + array + jnp.roll(array, +1, axis=1)
        return array * transparency_map, ()

    mask = jnp.zeros_like(transparency_map).at[tuple(origin)].set(1)

    view = jax.lax.scan(fin_diff, mask, None, radius)[0]

    # we now set a hard threshold > 0, but we can also think in the future
    # to use a cutoof at a different value to mimic the effect of a torch
    # (or eyesight for what matters)
    view = jnp.where(view > 0, 1, 0)

    # we add back the opaque tiles
    view = jnp.where(transparency_map == 0, 1, view)

    return view


def from_ascii_map(ascii_map: str, mapping: Dict[str, int] = {}) -> Array:
    """Converts an ASCII map into a 2D grid. The ASCII map is a string where each character
    represents a tile in the grid. The mapping dictionary can be used to map ASCII characters
    to integer values. By default, the mapping is as follows:
    - `#` is mapped to -1
    - `.` is mapped to 0
    
    Args:
        ascii_map (str): The ASCII map.
        mapping (Dict[str, int], optional): A dictionary mapping ASCII characters to integer \
        values. Defaults to {}.
    
    Returns:
        Array: A 2D grid representing the ASCII map."""
    mapping = {**{"#": -1, ".": 0}, **mapping}

    ascii_map = ascii_map.strip()
    max_width = max(len(line.strip()) for line in ascii_map.splitlines())
    grid = []
    for line in ascii_map.splitlines():
        line = line.strip()
        assert len(line) == max_width, "All lines must be the same length"
        row = [int(mapping.get(character, character)) for character in line]
        grid.append(row)

    return jnp.asarray(grid, dtype=jnp.int32)


class RoomsGrid(struct.PyTreeNode):
    """A grid of rooms. Each room is represented as a 2D grid of shape `(room_height, room_width)`,
    with walls set to -1 and empty tiles set to 0. The grid of rooms is represented as a 2D grid of
    shape `(rows * (room_height + 1), cols * (room_width + 1))`, with walls set to -1 and empty tiles
    set to 0. The grid of rooms is represented as a 2D grid of shape `(rows * (room_height + 1), cols * (room_width + 1))`,
    with walls set to -1 and empty tiles set to 0."""

    room_starts: Array  # shape (rows, cols)
    room_size: Tuple[int, int]

    @classmethod
    def create(
        cls, num_rows: int, num_cols: int, room_size: Tuple[int, int]
    ) -> RoomsGrid:
        """Creates a grid of rooms with the given number of rows and columns, and the given room size.

        Args:
            num_rows (int): The number of rows.
            num_cols (int): The number of columns.
            room_size (Tuple[int, int]): The size of each room `(height, width)`.

        Returns:
            RoomsGrid: A grid of rooms."""
        # generate rooms grid
        height = num_rows * (room_size[0] + 1)
        width = num_cols * (room_size[1] + 1)
        starts = jnp.mgrid[
            : height : room_size[0] + 1,
            : width : room_size[1] + 1,
        ].transpose(1, 2, 0)
        starts = jnp.asarray(starts, dtype=jnp.int32)
        sizes = jnp.ones((num_rows, num_cols, 2)) * jnp.asarray([[[room_size]]])
        sizes = jnp.asarray(sizes, dtype=jnp.int32)
        return cls(starts, room_size)

    def get_grid(self, occupied_positions: Array | None = None) -> Array:
        """Computes the array representation of the grid of rooms, with walls set to \
        -1 and empty tiles set to 0.
        
        Args:
            occupied_positions (Array, optional): A batch of extra occupied positions \
            of shape `(n, 2)`. Defaults to None.
        
        Returns:
            Array: A 2D grid of shape `(rows * (room_height + 1), cols * (room_width + 1))`."""
        room_size = self.room_size
        num_rows, num_cols = self.room_starts.shape[:2]
        grid = jnp.zeros(
            (1 + num_rows * (room_size[0] + 1), 1 + num_cols * (room_size[1] + 1))
        )
        grid = grid.at[jnp.arange(num_rows + 1) * (room_size[0] + 1)].set(-1)
        grid = grid.at[:, jnp.arange(num_cols + 1) * (room_size[1] + 1)].set(-1)

        if occupied_positions is not None:
            grid = grid.at[tuple(occupied_positions.T)].set(0)
        return grid

    def position_in_room(self, row: Array, col: Array, *, key: Array) -> Array:
        """Generates a random position in a given room.

        Args:
            row (Array): The row index of the room.
            col (Array): The column index of the room.
            key (Array): A random key.

        Returns:
            Array: A random position in the given room."""
        k1, k2 = jax.random.split(key)
        local_row = jax.random.randint(k1, (), minval=1, maxval=self.room_size[0])
        local_col = jax.random.randint(k2, (), minval=1, maxval=self.room_size[1])
        return jnp.asarray([local_row, local_col]) + self.room_starts[row, col]

    @partial(jax.jit, static_argnums=3)
    def position_on_border(
        self, row: Array, col: Array, side: int, *, key: Array
    ) -> Array:
        """Generates a random position on the border of a given room.
        Side is 0: west, 1: east, 2: north, 3: south (like padding)

        Args:
            row (Array): The row index of the room.
            col (Array): The column index of the room.
            side (int): The side of the room.
            key (Array): A random key.

        Returns:
            Array: A random position on the border of the given room."""
        starts = self.room_starts[row, col]
        room_size = self.room_size
        if side == 0:
            idx = jax.random.randint(key, (), minval=1, maxval=room_size[0] + 1)
            pos = (starts[0] + idx, starts[1])
        elif side == 1:
            idx = jax.random.randint(key, (), minval=1, maxval=room_size[0] + 1)
            pos = (starts[0] + idx, starts[1] + room_size[1] + 1)
        elif side == 2:
            idx = jax.random.randint(key, (), minval=1, maxval=room_size[1] + 1)
            pos = (starts[0], starts[1] + idx)
        elif side == 3:
            idx = jax.random.randint(key, (), minval=1, maxval=room_size[1] + 1)
            pos = (starts[0] + room_size[0] + 1, starts[1] + idx)
        else:
            raise ValueError("Side should be less than 4 and greater than -1")
        return jnp.asarray(pos)



import jax
import jax.numpy as jnp
import navix as nx

import gymnasium
import random
import time

from timeit import repeat

N_TIMEIT_LOOPS = 3
N_REPEAT = 5
N_TIMESTEPS = 1000
N_SEEDS = 10_000


def profile_navix_scan(seed):
    env = nx.environments.Room.create(
        height=10, width=5, max_steps=100, observation_fn=nx.observations.categorical
    )
    key = jax.random.PRNGKey(4)
    timestep = env._reset(key)
    actions = jax.random.randint(key, (N_TIMESTEPS,), 0, 6)

    timestep = jax.lax.scan(
        lambda carry, x: (env.step(carry, x), ()), timestep, actions
    )[0]

    return timestep


def profile_minigrid(seed):
    env = gymnasium.make("MiniGrid-Empty-16x16-v0", render_mode=None)
    observation, info = env.reset(seed=42)
    for _ in range(N_TIMESTEPS):
        action = random.randint(0, 4)
        observation, reward, terminated, truncated, info = env.step(action)

        if terminated or truncated:
            observation, info = env.reset()
    env.close()
    return observation


if __name__ == "__main__":
    # profile navix scanned
    print(
        "Profiling navix with `scan`, N_SEEDS = {}, N_TIMESTEPS = {}".format(
            N_SEEDS, N_TIMESTEPS
        )
    )
    seeds = jnp.arange(N_SEEDS)

    print(f"\tCompiling {profile_navix_scan}...")
    start = time.time()
    f_scan = jax.jit(jax.vmap(profile_navix_scan)).lower(seeds).compile()
    print("\tCompiled in {:.2f}s".format(time.time() - start))

    print("\tRunning ...")
    res_navix = repeat(
        lambda: f_scan(seeds).observation.block_until_ready(),
        number=N_TIMEIT_LOOPS,
        repeat=N_REPEAT,
    )
    res_navix = jnp.asarray(res_navix)
    print(f"\t {jnp.mean(res_navix)} ± {jnp.std(res_navix)}")

    # profile minigrid
    print("Profiling minigrid, N_SEEDS = 1, N_TIMESTEPS = {}".format(N_TIMESTEPS))
    res_minigrid = repeat(
        lambda: profile_minigrid(0), number=N_TIMEIT_LOOPS, repeat=N_REPEAT
    )
    res_minigrid = jnp.asarray(res_minigrid)
    print(f"\t {jnp.mean(res_minigrid)} ± {jnp.std(res_minigrid)}")



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import abc
from typing import Any, Callable, Dict, Tuple
import jax
import jax.numpy as jnp
from jax import Array
from flax import struct


from .. import rewards, terminations, observations, transitions
from ..rendering.cache import RenderingCache, TILE_SIZE
from ..states import State
from ..actions import DEFAULT_ACTION_SET
from ..spaces import Space, Discrete, Continuous


class StepType(struct.PyTreeNode):
    TRANSITION = jnp.asarray(0)
    """Standard timestep transition: the episode continues"""
    TRUNCATION = jnp.asarray(1)
    """The environment reached its maximum number of timesteps.
    The episode ended, but the agent could have still collected rewards.
    The value of the state is not 0"""
    TERMINATION = jnp.asarray(2)
    """The episode ended and the current state is an absorbing state."""


class Timestep(struct.PyTreeNode):
    t: Array
    """The number of timesteps elapsed from the last reset of the environment"""
    observation: Array
    """The observation corresponding to the current state (for POMDPs)"""
    action: Array
    """The action taken by the agent at the current timestep a_t = $\\pi(s_t)$, where $s_t$ is `state`"""
    reward: Array
    """The reward $r_{t=1}$ received by the agent after taking action $a_t$"""
    step_type: Array
    """The type of the current timestep, 0 for TRANSITION, 1 for TRUNCATION, 2 for TERMINATION"""
    state: State
    """The true state of the MDP, $s_t$ before taking action `action`"""
    info: Dict[str, Any] = struct.field(default_factory=dict)
    """Additional information about the environment. Useful for accumulations (e.g. returns)"""

    def is_truncation(self) -> Array:
        return self.step_type == StepType.TRUNCATION

    def is_termination(self) -> Array:
        return self.step_type == StepType.TERMINATION

    def is_transition(self) -> Array:
        return self.step_type == StepType.TRANSITION

    def is_done(self) -> Array:
        return jnp.logical_or(self.is_truncation(), self.is_termination())

    def is_start(self) -> Array:
        return self.t == 0


class Environment(struct.PyTreeNode):
    height: int = struct.field(pytree_node=False)
    width: int = struct.field(pytree_node=False)
    max_steps: int = struct.field(pytree_node=False)
    observation_space: Space = struct.field(pytree_node=False)
    action_space: Space = struct.field(pytree_node=False)
    reward_space: Space = struct.field(pytree_node=False)
    gamma: float = struct.field(pytree_node=False, default=0.99)
    penality_coeff: float = struct.field(pytree_node=False, default=0.0)
    observation_fn: Callable[[State], Array] = struct.field(
        pytree_node=False, default=observations.none
    )
    reward_fn: Callable[[State, Array, State], Array] = struct.field(
        pytree_node=False, default=rewards.DEFAULT_TASK
    )
    termination_fn: Callable[[State, Array, State], Array] = struct.field(
        pytree_node=False, default=terminations.DEFAULT_TERMINATION
    )
    transitions_fn: Callable[
        [State, Array, Tuple[Callable[[State], State], ...]], State
    ] = struct.field(pytree_node=False, default=transitions.DEFAULT_TRANSITION)
    action_set: Tuple[Callable[[State], State], ...] = struct.field(
        pytree_node=False, default=DEFAULT_ACTION_SET
    )

    @classmethod
    def create(
        cls,
        height: int,
        width: int,
        max_steps: int | None = None,
        observation_fn: Callable[[State], Array] = observations.symbolic,
        reward_fn: Callable[[State, Array, State], Array] = rewards.DEFAULT_TASK,
        termination_fn: Callable[
            [State, Array, State], Array
        ] = terminations.DEFAULT_TERMINATION,
        transitions_fn: Callable[
            [State, Array, Tuple[Callable[[State], State], ...]], State
        ] = transitions.DEFAULT_TRANSITION,
        action_set: Tuple[Callable[[State], State], ...] = DEFAULT_ACTION_SET,
        observation_space: Space | None = None,
        action_space: Space | None = None,
        reward_space: Space | None = None,
        **kwargs,
    ) -> Environment:

        if observation_space is None:
            observation_space = cls._get_obs_space_from_fn(
                width, height, observation_fn
            )
        if action_space is None:
            action_space = Discrete.create(len(action_set))
        if reward_space is None:
            reward_space = Continuous.create(
                shape=(), minimum=jnp.asarray(-1.0), maximum=jnp.asarray(1.0)
            )
        if max_steps is None:
            max_steps = 4 * height * width
        return cls(
            height=height,
            width=width,
            max_steps=max_steps,
            observation_fn=observation_fn,
            reward_fn=reward_fn,
            termination_fn=termination_fn,
            transitions_fn=transitions_fn,
            action_set=action_set,
            observation_space=observation_space,
            action_space=action_space,
            reward_space=reward_space,
            **kwargs,
        )

    @abc.abstractmethod
    def _reset(self, key: Array, cache: RenderingCache | None = None) -> Timestep:
        raise NotImplementedError()

    def reset(self, key: Array, cache: RenderingCache | None = None) -> Timestep:
        k1, k2 = jax.random.split(key)
        timestep = self._reset(k1, cache)
        timestep.info["return"] = jnp.asarray(0.0)
        return timestep.replace(state=timestep.state.replace(key=k2))

    def step(self, timestep: Timestep, action: Array) -> Timestep:
        # autoreset if necessary: 0 = transition, 1 = truncation, 2 = termination
        should_reset = timestep.step_type > 0
        return jax.lax.cond(
            should_reset,
            lambda timestep: self.reset(timestep.state.key, timestep.state.cache),
            lambda timestep: self._step(timestep, action),
            timestep,
        )

    def _step(self, timestep: Timestep, action: Array) -> Timestep:
        """
        Args:
            timestep (Timestep): The timestep at time $t$.
            action (Array): The action $a_t \\sim \\pi(A_t | s_t)$
        Returns:
            (Timestep): The timestep at time $t + 1$
        """
        # update agents
        state = self.transitions_fn(timestep.state, action, self.action_set)
        t = timestep.t + 1

        # calculate termination
        step_type = self.termination(timestep.state, action, state, timestep.t + 1)

        # calculate reward
        reward = self.reward_fn(timestep.state, action, state)
        reward = jax.lax.cond(
            step_type == StepType.TERMINATION,
            lambda reward: reward - self.penality_coeff * (t / self.max_steps),
            lambda reward: reward,
            reward,
        )

        new_timestep = Timestep(
            t=t,
            state=state,
            action=jnp.asarray(action),
            reward=reward,
            step_type=step_type,
            observation=self.observation_fn(state),
        )

        new_timestep.info["return"] = (
            timestep.info.get("return", jnp.asarray(0.0)) + reward
        )

        # build timestep
        return new_timestep

    def termination(
        self, prev_state: State, action: Array, state: State, t: Array
    ) -> Array:
        terminated = self.termination_fn(prev_state, action, state)
        truncated = t >= self.max_steps
        return terminations.check_truncation(terminated, truncated)

    @staticmethod
    def _get_obs_space_from_fn(
        width: int, height: int, observation_fn: Callable[[State], Array]
    ) -> Space:
        if observation_fn == observations.none:
            return Continuous.create(
                shape=(), minimum=jnp.asarray(0.0), maximum=jnp.asarray(0.0)
            )
        elif observation_fn == observations.categorical:
            return Discrete.create(n_elements=9, shape=(height, width))
        elif observation_fn == observations.categorical_first_person:
            radius = observations.RADIUS
            return Discrete.create(n_elements=9, shape=(radius + 1, radius * 2 + 1))
        elif observation_fn == observations.rgb:
            return Discrete.create(
                256,
                shape=(height * TILE_SIZE, width * TILE_SIZE, 3),
                dtype=jnp.uint8,
            )
        elif observation_fn == observations.rgb_first_person:
            radius = observations.RADIUS
            return Discrete.create(
                n_elements=256,
                shape=(radius * TILE_SIZE + 1, radius * TILE_SIZE * 2 + 1, 3),
                dtype=jnp.uint8,
            )
        elif observation_fn == observations.symbolic:
            return Discrete.create(
                n_elements=9,
                shape=(height, width, 3),
                dtype=jnp.uint8,
            )
        elif observation_fn == observations.symbolic_first_person:
            radius = observations.RADIUS
            return Discrete.create(
                n_elements=256,
                shape=(radius + 1, radius * 2 + 1, 3),
                dtype=jnp.uint8,
            )
        else:
            raise NotImplementedError(
                "Unknown observation space for observation function {}".format(
                    observation_fn
                )
            )



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import Union

import jax
import jax.numpy as jnp
from jax import Array

from navix import observations, rewards, terminations

from ..components import EMPTY_POCKET_ID
from ..rendering.cache import RenderingCache
from . import Environment
from ..entities import Player, Goal, Lava
from ..states import State
from . import Timestep
from ..grid import room
from .registry import register_env


class LavaGap(Environment):
    def _reset(self, key: Array, cache: Union[RenderingCache, None] = None) -> Timestep:
        # check minimum height and width
        assert (
            self.height > 3
        ), f"Room height must be greater than 3, got {self.height} instead"
        assert (
            self.width > 4
        ), f"Room width must be greater than 5, got {self.width} instead"

        key, k1, k2 = jax.random.split(key, num=3)

        grid = room(height=self.height, width=self.width)

        # player
        player_pos = jnp.asarray([1, 1])
        player_dir = jnp.asarray(0)
        player = Player.create(
            position=player_pos, direction=player_dir, pocket=EMPTY_POCKET_ID
        )
        # goal
        goal_pos = jnp.asarray([self.height - 2, self.width - 2])
        goals = Goal.create(position=goal_pos, probability=jnp.asarray(1.0))

        # lava positions
        gap_row = jax.random.randint(k1, (), 1, self.height - 1)  # col

        col = jax.random.randint(k2, (), minval=2, maxval=self.width - 2)
        lava_row = jnp.arange(1, self.height - 1)
        lava_cols = jnp.asarray([col] * (self.height - 2))
        lava_pos = jnp.stack((lava_row, lava_cols), axis=1)
        # remove lava where the door is
        lava_pos = jnp.delete(lava_pos, gap_row - 1, axis=0, assume_unique_indices=True)
        lavas = Lava.create(position=lava_pos)

        entities = {
            "player": player[None],
            "goal": goals[None],
            "lava": lavas,
        }

        state = State(
            key=key,
            grid=grid,
            cache=cache or RenderingCache.init(grid),
            entities=entities,
        )
        return Timestep(
            t=jnp.asarray(0, dtype=jnp.int32),
            observation=self.observation_fn(state),
            action=jnp.asarray(-1, dtype=jnp.int32),
            reward=jnp.asarray(0.0, dtype=jnp.float32),
            step_type=jnp.asarray(0, dtype=jnp.int32),
            state=state,
        )


register_env(
    "Navix-LavaGapS5-v0",
    lambda *args, **kwargs: LavaGap.create(
        *args,
        **kwargs,
        height=5,
        width=5,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
    ),
)
register_env(
    "Navix-LavaGapS6-v0",
    lambda *args, **kwargs: LavaGap.create(
        *args,
        **kwargs,
        height=6,
        width=6,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
    ),
)
register_env(
    "Navix-LavaGapS7-v0",
    lambda *args, **kwargs: LavaGap.create(
        *args,
        **kwargs,
        height=7,
        width=7,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
    ),
)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import Union
import jax
import jax.numpy as jnp
from jax import Array
from flax import struct

from navix import observations, rewards, terminations

from ..components import EMPTY_POCKET_ID
from ..rendering.cache import RenderingCache
from . import Environment
from ..entities import Player, Goal, Lava
from ..states import State
from . import Timestep
from .registry import register_env


class Crossings(Environment):
    n_crossings: int = struct.field(pytree_node=False, default=1)
    is_lava: bool = struct.field(pytree_node=False, default=False)

    def _reset(self, key: Array, cache: Union[RenderingCache, None] = None) -> Timestep:
        assert (
            self.height == self.width
        ), f"Crossings are only defined for square grids, got height {self.height} and \
            width {self.width}"
        # check minimum height and width
        key, k1, k2 = jax.random.split(key, num=3)

        grid = jnp.zeros((self.height - 2, self.width - 2), dtype=jnp.int32)

        # player
        player_pos = jnp.asarray([1, 1])
        player_dir = jnp.asarray(0)
        player = Player.create(
            position=player_pos, direction=player_dir, pocket=EMPTY_POCKET_ID
        )
        # goal
        goal_pos = jnp.asarray([self.height - 2, self.width - 2])
        goals = Goal.create(position=goal_pos, probability=jnp.asarray(1.0))

        entities = {
            "player": player[None],
            "goal": goals[None],
        }

        # crossings
        obstacles_hor = jnp.mgrid[
            1 : self.height - 2 : 2, 1 : self.width - 1
        ].transpose(1, 2, 0)
        obstacles_ver = jnp.mgrid[
            1 : self.height - 1, 1 : self.width - 2 : 2
        ].transpose(2, 1, 0)
        all_obstacles_pos = jnp.concatenate([obstacles_hor, obstacles_ver])
        num_obstacles = min(self.n_crossings, len(all_obstacles_pos))
        obstacles_pos = jax.random.choice(
            k1, all_obstacles_pos, (num_obstacles,), replace=False
        )

        if self.is_lava:
            entities["lava"] = Lava.create(position=obstacles_pos)
        else:
            grid = grid.at[tuple(obstacles_pos.T)].set(-1)

        # path to goal
        def update(direction, start, grid, step_size):
            return jax.lax.cond(
                direction == jnp.asarray(0, dtype=jnp.int32),
                lambda: (
                    start + jnp.asarray([0, step_size]),
                    jax.lax.dynamic_update_slice(
                        grid, jnp.zeros((1, step_size), dtype=jnp.int32), tuple(start.T)
                    ),
                ),
                lambda: (
                    start + jnp.asarray([step_size, 0]),
                    jax.lax.dynamic_update_slice(
                        grid, jnp.zeros((step_size, 1), dtype=jnp.int32), tuple(start.T)
                    ),
                ),
            )

        start = jnp.asarray([0, 0], dtype=jnp.int32)
        step_size = 3
        for i in range(10):
            k2, k3 = jax.random.split(k2)
            direction = jax.random.randint(k2, (), minval=0, maxval=2)
            start, grid = update(direction, start, grid, step_size)

        grid = jnp.pad(grid, 1, mode="constant", constant_values=-1)

        state = State(
            key=key,
            grid=grid,
            cache=RenderingCache.init(grid),
            entities=entities,
        )
        return Timestep(
            t=jnp.asarray(0, dtype=jnp.int32),
            observation=self.observation_fn(state),
            action=jnp.asarray(-1, dtype=jnp.int32),
            reward=jnp.asarray(0.0, dtype=jnp.float32),
            step_type=jnp.asarray(0, dtype=jnp.int32),
            state=state,
        )


register_env(
    "Navix-SimpleCrossingsS9N1-v0",
    lambda *args, **kwargs: Crossings.create(
        height=9,
        width=9,
        n_crossings=1,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-SimpleCrossingsS9N2-v0",
    lambda *args, **kwargs: Crossings.create(
        height=9,
        width=9,
        n_crossings=2,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-SimpleCrossingsS9N3-v0",
    lambda *args, **kwargs: Crossings.create(
        height=9,
        width=9,
        n_crossings=3,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-SimpleCrossingsS11N5-v0",
    lambda *args, **kwargs: Crossings.create(
        height=11,
        width=11,
        n_crossings=5,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        *args,
        **kwargs,
    ),
)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations
from typing import Union

import jax
import jax.numpy as jnp
from jax import Array
from flax import struct

from navix import observations, rewards, terminations

from ..components import EMPTY_POCKET_ID
from ..entities import Entities, Goal, Player, Ball
from ..states import State
from ..grid import random_positions, random_directions, room
from ..rendering.cache import RenderingCache
from ..rendering.registry import PALETTE
from .environment import Environment, Timestep
from .registry import register_env


class DynamicObstacles(Environment):
    random_start: bool = struct.field(pytree_node=False, default=False)
    n_obstacles: int = struct.field(pytree_node=False, default=2)

    def _reset(self, key: Array, cache: Union[RenderingCache, None] = None) -> Timestep:
        key, k1, k2, k3 = jax.random.split(key, 4)

        # map
        grid = room(height=self.height, width=self.width)

        # goal and player
        if self.random_start:
            player_pos = random_positions(k1, grid)
            direction = random_directions(k2, n=1)
        else:
            player_pos = jnp.asarray([1, 1])
            direction = jnp.asarray(0)
        # player
        player = Player.create(
            position=player_pos,
            direction=direction,
            pocket=EMPTY_POCKET_ID,
        )
        # goal
        goal_pos = jnp.asarray([self.height - 2, self.width - 2])
        goal = Goal.create(position=goal_pos, probability=jnp.asarray(1.0))

        # balls
        exclude = jnp.stack([player_pos, goal_pos])
        ball_pos = random_positions(k3, grid, n=self.n_obstacles, exclude=exclude)
        balls = Ball.create(
            position=ball_pos,
            colour=jnp.tile(PALETTE.BLUE, (self.n_obstacles,)),
            probability=jnp.ones(self.n_obstacles),
        )

        entities = {
            Entities.PLAYER: player[None],
            Entities.GOAL: goal[None],
            Entities.BALL: balls,
        }

        # systems
        state = State(
            key=key,
            grid=grid,
            cache=cache or RenderingCache.init(grid),
            entities=entities,
        )

        return Timestep(
            t=jnp.asarray(0, dtype=jnp.int32),
            observation=self.observation_fn(state),
            action=jnp.asarray(0, dtype=jnp.int32),
            reward=jnp.asarray(0.0, dtype=jnp.float32),
            step_type=jnp.asarray(0, dtype=jnp.int32),
            state=state,
        )


register_env(
    "Navix-Dynamic-Obstacles-5x5-v0",
    lambda *args, **kwargs: DynamicObstacles.create(
        height=5,
        width=5,
        n_obstacles=2,
        random_start=False,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-Dynamic-Obstacles-5x5-Random-v0",
    lambda *args, **kwargs: DynamicObstacles.create(
        height=5,
        width=5,
        n_obstacles=2,
        random_start=True,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-Dynamic-Obstacles-6x6-v0",
    lambda *args, **kwargs: DynamicObstacles.create(
        height=6,
        width=6,
        n_obstacles=3,
        random_start=False,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-Dynamic-Obstacles-6x6-Random-v0",
    lambda *args, **kwargs: DynamicObstacles.create(
        height=6,
        width=6,
        n_obstacles=3,
        random_start=True,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-Dynamic-Obstacles-8x8-v0",
    lambda *args, **kwargs: DynamicObstacles.create(
        height=8,
        width=8,
        n_obstacles=4,
        random_start=False,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-Dynamic-Obstacles-16x16-v0",
    lambda *args, **kwargs: DynamicObstacles.create(
        height=16,
        width=16,
        n_obstacles=8,
        random_start=False,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        *args,
        **kwargs,
    ),
)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations

import os
from PIL import Image

import jax
from jax import Array
import jax.numpy as jnp


SPRITES_DIR = os.path.normpath(
    os.path.join(__file__, "..", "..", "..", "assets", "sprites")
)
MIN_TILE_SIZE = 8
TILE_SIZE = MIN_TILE_SIZE


def load_sprite(name: str) -> Array:
    """Loads an image from disk in RGB space.
    Args:
        path(str): the filepath of the image on disk

    Returns:
        (Array): a jax.Array of shape (H, W, C)"""
    path = os.path.join(SPRITES_DIR, f"{name}.png")
    image = Image.open(path)
    array = jnp.asarray(image)
    resized = jax.image.resize(array, (TILE_SIZE, TILE_SIZE, 3), method="cubic")
    return jnp.asarray(resized, dtype=jnp.uint8)


class PALETTE:
    RED: Array = jnp.asarray(0, dtype=jnp.uint8)
    GREEN: Array = jnp.asarray(1, dtype=jnp.uint8)
    BLUE: Array = jnp.asarray(2, dtype=jnp.uint8)
    PURPLE: Array = jnp.asarray(3, dtype=jnp.uint8)
    YELLOW: Array = jnp.asarray(4, dtype=jnp.uint8)
    GREY: Array = jnp.asarray(5, dtype=jnp.uint8)
    UNSET: Array = jnp.asarray(255, dtype=jnp.uint8)

    @classmethod
    def as_string(cls):
        return ["red", "green", "blue", "purple", "yellow", "grey"]

    @classmethod
    def as_array(cls):
        return [cls.RED, cls.GREEN, cls.BLUE, cls.PURPLE, cls.YELLOW, cls.GREY]


class SpritesRegistry:
    def __init__(self):
        self.registry = {}
        self.build_registry()

    def build_registry(self):
        """Populates the sprites registry for all entities."""
        self.set_wall_sprite()
        self.set_floor_sprite()
        self.set_goal_sprite()
        self.set_key_sprite()
        self.set_player_sprite()
        self.set_door_sprite()
        self.set_lava_sprite()
        self.set_ball_sprite()
        self.set_box_sprite()

    def set_wall_sprite(self):
        self.registry["wall"] = load_sprite("wall")

    def set_floor_sprite(self):
        self.registry["floor"] = load_sprite("floor")

    def set_goal_sprite(self):
        self.registry["goal"] = load_sprite("goal")

    def set_key_sprite(self):
        keys_coloured = [
            load_sprite("key" + f"_{colour}") for colour in PALETTE.as_string()
        ]
        self.registry["key"] = jnp.stack(keys_coloured, axis=0)

    def set_player_sprite(self):
        self.registry["player"] = jnp.stack(
            [
                load_sprite("player_east"),
                load_sprite("player_south"),
                load_sprite("player_west"),
                load_sprite("player_north"),
            ]
        )

    def set_door_sprite(self):
        door = jnp.zeros(
            (len(PALETTE.as_string()), 3, TILE_SIZE, TILE_SIZE, 3), dtype=jnp.uint8
        )
        for c_idx, colour in enumerate(PALETTE.as_string()):
            for s_idx, state in enumerate(["closed", "open", "locked"]):
                sprite = load_sprite("door" + f"_{state}" + f"_{colour}")
                door = door.at[c_idx, s_idx].set(sprite)
        self.registry["door"] = door

    def set_lava_sprite(self):
        self.registry["lava"] = load_sprite("lava")

    def set_ball_sprite(self):
        ball_coloured = [
            load_sprite("ball" + f"_{colour}") for colour in PALETTE.as_string()
        ]
        self.registry["ball"] = jnp.stack(ball_coloured, axis=0)

    def set_box_sprite(self):
        box_coloured = [
            load_sprite("box" + f"_{colour}") for colour in PALETTE.as_string()
        ]
        self.registry["box"] = jnp.stack(box_coloured, axis=0)


# initialise sprites registry
SPRITES_REGISTRY = SpritesRegistry().registry



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations
import jax

import jax.numpy as jnp
from jax import Array

from .rendering.cache import TILE_SIZE, unflatten_patches
from .components import DISCARD_PILE_IDX, Directional, HasColour, Openable
from .states import State
from .grid import align, idx_from_coordinates, crop, view_cone
from .entities import EntityIds


RADIUS = 3


def none(state: State) -> Array:
    """An empty observation represented as an array of shape f32[0].
    Useful for testing purposes.

    Args:
        state (State): The current state of the game.

    Returns:
        Array: A 0-shaped array `f32[0]`."""
    return jnp.asarray(())


def categorical(state: State) -> Array:
    """Fully observable grid with a categorical state representation.
    Each entity is represented by its unique integer tag.
    
    Args:
        state (State): The current state of the game.
    
    Returns:
        Array: A grid of integers, where each integer represents an entity, \
        represented as an array of shape `i32[H, W]`, where `H` and `W` are the height \
        and width of the grid."""
    # get idx of entity on the set of patches
    indices = idx_from_coordinates(state.grid, state.get_positions())
    # get tags corresponding to the entities
    tags = state.get_tags()
    # set tags on the flat set of patches
    shape = state.grid.shape
    grid = state.grid.reshape(-1).at[indices].set(tags)
    # unflatten patches to reconstruct the grid
    return grid.reshape(shape)


def categorical_first_person(state: State) -> Array:
    """Categorical state representation, but cropped to the agent's view, and aligned \
    with the agent's direction, such that the agent always points upwards.
    
    Args:
        state (State): The current state of the game.

    Returns:
        Array: A grid of integers, where each integer represents an entity, \
        represented as an array of shape `i32[2 * RADIUS + 1, 2 * RADIUS + 1]`."""
    # get transparency map
    transparency_map = jnp.where(state.grid == 0, 1, 0)
    positions = state.get_positions()
    transparent = state.get_transparency()
    transparency_map = transparency_map.at[tuple(positions.T)].set(~transparent)

    # apply view mask
    player = state.get_player()
    view = view_cone(transparency_map, player.position, RADIUS)

    # get categorical representation
    tags = state.get_tags()
    obs = state.grid.at[tuple(positions.T)].set(tags) * view

    # crop grid to agent's view
    obs = crop(obs, player.position, player.direction, RADIUS)

    return obs


def symbolic(state: State) -> Array:
    """Fully observable grid with a symbolic state representation as originally \
    proposed in the MiniGrid environment.
    The symbol is a triple of (OBJECT_TAG, COLOUR_IDX, OPEN/CLOSED/LOCKED). The
    last layer might also contain the direction of the entity, for example, the
    direction of the agent.
    
    Args:
        state (State): The current state of the game.
        
    Returns:
        Array: A grid of integers, where each integer represents an entity, \
        represented as an array of shape `u8[H, W, 3]`, where `H` and `W` are the height \
        and width of the grid."""
    # initialise as all floors
    H, W = state.grid.shape
    obs = jnp.zeros((H, W, 3), dtype=jnp.uint8)
    wall_symbol = jnp.array([EntityIds.WALL, 5, 0], dtype=jnp.uint8)
    floor_symbol = jnp.array([EntityIds.FLOOR, 0, 0], dtype=jnp.uint8)
    obs = jnp.where(state.grid[..., None] == -1, wall_symbol, floor_symbol)

    # place entities
    for entity_class in state.entities:
        entity = state.entities[entity_class]
        tag = entity.tag
        # colour layer
        if isinstance(entity, HasColour):
            colour = entity.colour
        else:
            colour = jnp.zeros(entity.shape)
        # state layer
        if isinstance(entity, Openable):
            entity_state = entity.open + (entity.requires != jnp.zeros(entity.shape))
        elif isinstance(entity, Directional):
            entity_state = entity.direction
        else:
            entity_state = jnp.zeros(entity.shape)
        entity_symbol = jnp.stack([tag, colour, entity_state], axis=-1, dtype=jnp.uint8)
        obs = obs.at[tuple(entity.position.T)].set(entity_symbol)
    return obs


def symbolic_first_person(state: State) -> Array:
    """First person view with a symbolic state representation, but cropped to the \
    agent's view, and aligned with the agent's direction, such that the agent always \
    points upwards. See `symbolic` for more details.
    
    Args:
        state (State): The current state of the game.
    
    Returns:
        Array: A grid of integers, where each integer represents an entity, \
        represented as an array of shape `u8[2 * RADIUS + 1, 2 * RADIUS + 1, 3]`."""
    # get transparency map
    obs = symbolic(state)

    # replace player with pocket to show them what they are carrying
    player = state.get_player()
    obs = obs.at[tuple(player.position.T)].set(
        jnp.asarray([EntityIds.FLOOR, 0, 0], dtype=jnp.uint8)
    )

    # crop to first person view
    obs = crop(
        obs,
        player.position,
        player.direction,
        RADIUS,
        padding_value=255,
    )
    # replace padding symbol with walls
    wall_symbol = jnp.array([EntityIds.WALL, 5, 0], dtype=jnp.uint8)
    obs = jnp.where(obs == 255, wall_symbol, obs)
    return obs


def rgb(state: State) -> Array:
    """Fully observable grid with an RGB state representation.
    Each entity is represented by its unique RGB sprite. The RGB sprites are \
    stored in a cache, and the entities are placed on the grid according to their \
    positions.
    
    Args:
        state (State): The current state of the game.
    
    Returns:
        Array: An RGB image of the grid, represented as an array of shape \
        `u8[H * S, W * S, 3]`, where `H` and `W` are the height and width of the grid,
        and `S` is the size of the tile."""
    # get idx of entity on the flat set of patches
    indices = idx_from_coordinates(state.grid, state.get_positions())
    # get tiles corresponding to the entities
    tiles = state.get_sprites()
    # set tiles on the flat set of patches
    patches = state.cache.patches.at[indices].set(tiles)
    # remove discard pile
    patches = patches[:DISCARD_PILE_IDX]
    # unflatten patches to reconstruct the image
    image_size = (
        state.grid.shape[0] * TILE_SIZE,
        state.grid.shape[1] * TILE_SIZE,
    )
    image = unflatten_patches(patches, image_size)
    return image


def rgb_first_person(state: State) -> Array:
    """First person view with an RGB state representation.
    The image is cropped to the agent's view, and aligned with the agent's direction, \
    such that the agent always points upwards. See `rgb` for more details.
    See `rgb` for more details.

    Args:
        state (State): The current state of the game.
    
    Returns:
        Array: An RGB image of the agent's view, represented as an array of shape \
        `u8[(2 * RADIUS + 1) * S, (2 * RADIUS + 1) * S, 3]`, where 
        `S` is the size of the tile."""
    # calculate final image size
    # get agent's view
    # image_size = (
    #     state.grid.shape[0] * TILE_SIZE,
    #     state.grid.shape[1] * TILE_SIZE,
    # )
    # transparency_map = jnp.where(state.grid == 0, 1, 0)
    # positions = state.get_positions()
    # transparent = state.get_transparency()
    # transparency_map = transparency_map.at[tuple(positions.T)].set(~transparent)
    # view = view_cone(transparency_map, player.position, RADIUS)
    # view = jax.image.resize(view, image_size, method="nearest")
    # view = jnp.tile(view[..., None], (1, 1, 3))

    # get sprites aligned to player's direction
    sprites = state.get_sprites()
    sprites = jax.vmap(lambda x: align(x, jnp.asarray(0), player.direction))(sprites)

    # align sprites to player's direction
    indices = idx_from_coordinates(state.grid, state.get_positions())
    patches = state.cache.patches.at[indices].set(sprites)

    # remove discard pile
    patches = patches[:DISCARD_PILE_IDX]
    # rearrange the sprites in a grid
    patchwork = patches.reshape(*state.grid.shape, *patches.shape[1:])

    # crop grid to agent's view
    player = state.get_player()
    patchwork = crop(patchwork, player.position, player.direction, RADIUS)

    # reconstruct image
    obs = jnp.swapaxes(patchwork, 1, 2)
    shape = obs.shape
    obs = obs.reshape(shape[0] * shape[1], shape[2] * shape[3], *shape[4:])
    return obs



from typing import Callable
from jax import Array
from flax import struct

from . import rewards, terminations
from .states import State


class Task(struct.PyTreeNode):
    """Task formuation as described in https://arxiv.org/abs/1609.01995"""

    reward_fn: Callable[[State, Array, State], Array]
    termination_fn: Callable[[State, Array, State], Array]


NAVIGATION = Task(rewards.on_goal_reached, terminations.on_goal_reached)

GO_TO_DOOR = Task(terminations.on_door_done, terminations.on_door_done)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""The *action* system determines the next state of the environment \
given the current state and an action."""


from __future__ import annotations
from typing import Tuple

import jax
from jax import Array
import jax.numpy as jnp

from .entities import Entities, Player
from .states import EventsManager, State
from .components import DISCARD_PILE_COORDS, Pickable
from .grid import translate, rotate, positions_equal


def _rotate(state: State, spin: int) -> State:
    if Entities.PLAYER not in state.entities:
        return state

    player = state.get_player(idx=0)

    # update player's direction
    direction = rotate(player.direction, spin)

    # update sprite representation
    player = player.replace(direction=direction)

    state = state.set_player(player)

    return state


def _can_walk_there(state: State, position: Array) -> Tuple[Array, EventsManager]:
    # according to the grid
    walkable = jnp.equal(state.grid[tuple(position)], 0)
    events = jax.lax.cond(
        walkable,
        lambda: state.events,
        lambda: state.events.record_grid_hit(position),
    )

    for k in state.entities:
        same_position = positions_equal(state.entities[k].position, position)
        events = jax.lax.cond(
            jnp.any(same_position),
            lambda x: x.record_walk_into(state.entities[k], position),
            lambda x: x,
            events,
        )
        obstructs = jnp.logical_and(
            jnp.logical_not(state.entities[k].walkable), same_position
        )
        walkable = jnp.logical_and(walkable, jnp.any(jnp.logical_not(obstructs)))
    return jnp.asarray(walkable, dtype=jnp.bool_), events


def _move(state: State, direction: Array) -> State:
    if Entities.PLAYER not in state.entities:
        return state

    player = state.get_player(idx=0)
    new_position = translate(player.position, direction)
    can_move, events = _can_walk_there(state, new_position)
    new_position = jnp.where(can_move, new_position, player.position)
    # update structs
    player = player.replace(position=new_position)
    state = state.set_player(player)
    return state.replace(events=events)


def noop(state: State) -> State:
    """No operation. Does nothing.

    Args:
        state (State): The current state.
    
    Returns:
        State: The same state."""
    return state


def rotate_cw(state: State) -> State:
    """Rotates the player clockwise.
    
    Args:
        state (State): The current state.
    
    Returns:
        State: The new state with the player rotated clockwise."""
    return _rotate(state, 1)


def rotate_ccw(state: State) -> State:
    """Rotates the player counter-clockwise.
    
    Args:
        state (State): The current state.

    Returns:
        State: The new state with the player rotated counter-clockwise."""
    return _rotate(state, -1)


def forward(state: State) -> State:
    """Moves the player forward.
    
    Args:
        state: The current state.

    Returns:
        State: The new state with the player moved forward."""
    player = state.get_player(idx=0)
    return _move(state, player.direction)


def right(state: State) -> State:
    """Steps the player to the right without changing the direction.

    Args:
        state (State): The current state.

    Returns:
        State: The new state with the player moved to the right."""
    player = state.get_player(idx=0)
    return _move(state, player.direction + 1)


def backward(state: State) -> State:
    """Steps the player backward without changing the direction.
        
        Args:
            state (State): The current state.

        Returns:
            State: The new state with the player moved backward."""
    player = state.get_player(idx=0)
    return _move(state, player.direction + 2)


def left(state: State) -> State:
    """Steps the player to the left without changing the direction.

    Args:
        state (State): The current state.
    
    Returns:
        State: The new state with the player moved to the left."""
    player = state.get_player(idx=0)
    return _move(state, player.direction + 3)


def pickup(state: State) -> State:
    """Picks up an item in front of the player and puts it in the pocket.
    Args:
        state (State): The current state.
    Returns:
        State: The new state with the player entity having the item in the pocket."""
    if Entities.KEY not in state.entities:
        return state

    player = state.get_player(idx=0)
    keys = state.get_keys()

    position_in_front = translate(player.position, player.direction)

    key_found = positions_equal(position_in_front, keys.position)

    # update keys
    positions = jnp.where(key_found, DISCARD_PILE_COORDS, keys.position)
    keys = keys.replace(position=positions)

    # update player's pocket, if the pocket has something else, we overwrite it
    key = jnp.sum(keys.id * key_found, dtype=jnp.int32)
    player = jax.lax.cond(
        jnp.any(key_found), lambda: player.replace(pocket=key), lambda: player
    )

    # update events
    events = jax.lax.cond(
        jnp.any(key_found),
        lambda: state.events.record_key_pickup(keys, position_in_front),
        lambda: state.events,
    )

    state = state.set_player(player)
    state = state.set_keys(keys)
    state = state.set_events(events)
    return state


def drop(state: State) -> State:
    """Replaces the position in front of the player with the item in the pocket.

    Args:
        state (State): The current state.
    
    Returns:
        State: The new state with the item in the pocket dropped in front of the player."""
    player = state.get_player(idx=0)

    position_in_front = translate(player.position, player.direction)

    has_item = player.pocket != -1
    can_drop, events = _can_walk_there(state, position_in_front)
    can_drop = jnp.logical_and(can_drop, has_item)

    for k in state.entities:
        entity = state.entities[k]
        if isinstance(entity, Pickable):
            cond = jnp.logical_and(can_drop, entity.position == DISCARD_PILE_COORDS)
            position = jnp.where(cond, position_in_front, entity.position)
            entity = entity.replace(position=position)
            state.set_entity(k, entity)
    return state


def toggle(state: State) -> State:
    """Toggles an openable object (like a door) if possible.

    Args:
        state (State): The current state.
    
    Returns:
        State: The new state with the openable object toggled."""
    return open(state)


def open(state: State) -> State:
    """Unlocks and opens an openable object (like a door) if possible.
    
    Args:
        state (State): The current state.
    
    Returns:
        State: The new state with the openable object opened."""
    if Entities.DOOR not in state.entities:
        return state

    # get the tile in front of the player
    player = state.get_player(idx=0)
    doors = state.get_doors()

    position_in_front = translate(player.position, player.direction)

    # check if there is a door in front of the player
    door_found = positions_equal(position_in_front, doors.position)

    # and that, if so, either it does not require a key or the player has the key
    locked = doors.requires != -1
    key_match = player.pocket == doors.requires
    can_open = door_found & (key_match | ~locked)

    # update doors if closed and can_open
    do_open = ~doors.open & can_open
    open = jnp.where(do_open, True, doors.open)
    requires = jnp.where(do_open, -1, doors.requires)
    doors = doors.replace(open=open, requires=requires)

    # remove key from player's pocket
    pocket = jnp.asarray(player.pocket * jnp.any(can_open), dtype=jnp.int32)
    player = jax.lax.cond(
        jnp.any(can_open), lambda: player.replace(pocket=pocket), lambda: player
    )

    # update events
    events = jax.lax.cond(
        jnp.any(do_open),
        lambda: state.events.record_door_opening(doors, position_in_front),
        lambda: state.events,
    )

    state = state.set_player(player)
    state = state.set_doors(doors)
    state = state.set_events(events)

    return state


def done(state: State) -> State:
    """A placeholder action that does nothing, but is a signal to the environment that the episode is over.
    This action does not terminate the episode, unless the termination function explicitly checks for it (not default).
    
    Args:
        state (State): The current state.
    
    Returns:
        State: The same state."""
    return state


# DEFAULT_ACTION_SET = (
#     rotate_ccw,
#     rotate_cw,
#     forward,
#     pickup,
#     drop,
#     toggle,
#     done
# )
"""Default action set from Minigrid. See
https://github.com/Farama-Foundation/Minigrid/blob/master/minigrid/core/actions.py"""


COMPLETE_ACTION_SET = (
    noop,
    rotate_cw,
    rotate_ccw,
    forward,
    right,
    backward,
    left,
    pickup,
    open,
    done,
)
"""Complete action set for the environment.
This set includes all the actions that can be taken by the agent, and does not mirror the Minigrid action set."""

MINIGRID_ACTION_SET = (
    rotate_ccw,
    rotate_cw,
    forward,
    pickup,
    drop,
    toggle,
    done,
)
"""Default action set from Minigrid. See
https://github.com/Farama-Foundation/Minigrid/blob/master/minigrid/core/actions.py"""

DEFAULT_ACTION_SET = MINIGRID_ACTION_SET



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations
from typing import Union

import jax
import jax.numpy as jnp
from jax import Array

from navix import observations, rewards, terminations

from ..components import EMPTY_POCKET_ID
from ..entities import Entities, Goal, Player, Wall
from ..states import State
from ..grid import (
    random_positions,
    random_directions,
    room,
    horizontal_wall,
    vertical_wall,
)
from ..rendering.cache import RenderingCache
from .environment import Environment, Timestep
from .registry import register_env


class FourRooms(Environment):
    def _reset(self, key: Array, cache: Union[RenderingCache, None] = None) -> Timestep:
        assert self.height > 4, f"Insufficient height for room {self.height} < 4"
        assert self.width > 4, f"Insufficient width for room {self.width} < 4"
        key, k1, k2 = jax.random.split(key, 3)

        # map
        grid = room(height=self.height, width=self.width)

        # vertical partition
        opening_1 = jax.random.randint(k1, shape=(), minval=1, maxval=self.height // 2)
        opening_2 = jax.random.randint(
            k1, shape=(), minval=self.height // 2 + 2, maxval=self.height
        )
        openings = jnp.stack([opening_1, opening_2])
        wall_pos_vert = vertical_wall(grid, 9, openings)

        # horizontal partition
        opening_1 = jax.random.randint(k2, shape=(), minval=1, maxval=self.width // 2)
        opening_2 = jax.random.randint(
            k1, shape=(), minval=self.width // 2 + 2, maxval=self.width
        )
        openings = jnp.stack([opening_1, opening_2])
        wall_pos_hor = horizontal_wall(grid, 9, openings)

        walls_pos = jnp.concatenate([wall_pos_vert, wall_pos_hor])
        walls = Wall.create(position=walls_pos)

        # player
        player_pos, goal_pos = random_positions(k1, grid, n=2, exclude=walls_pos)
        direction = random_directions(k2, n=1)
        player = Player.create(
            position=player_pos,
            direction=direction,
            pocket=EMPTY_POCKET_ID,
        )
        # goal
        goal = Goal.create(position=goal_pos, probability=jnp.asarray(1.0))

        entities = {
            Entities.PLAYER: player[None],
            Entities.GOAL: goal[None],
            Entities.WALL: walls,
        }

        # systems
        state = State(
            key=key,
            grid=grid,
            cache=cache or RenderingCache.init(grid),
            entities=entities,
        )

        return Timestep(
            t=jnp.asarray(0, dtype=jnp.int32),
            observation=self.observation_fn(state),
            action=jnp.asarray(0, dtype=jnp.int32),
            reward=jnp.asarray(0.0, dtype=jnp.float32),
            step_type=jnp.asarray(0, dtype=jnp.int32),
            state=state,
        )


register_env(
    "Navix-FourRooms-v0",
    lambda *args, **kwargs: FourRooms.create(
        height=19,
        width=19,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        *args,
        **kwargs,
    ),
)
